/*
 * job.c
 *
 * Copyright (C) 1997,98 Rasca, Berlin
 * Copyright (C) 2003,04 Karl H. Beckers, Frankfurt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * This file defines the job data structure and some functions to
 * set/get info about a job (kinda tries to be OO ;S )
 *
 */

#include "../config.h" 	/* autoconf output */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h> /* PATH_MAX */
#include <X11/Intrinsic.h>
#ifdef HAVE_LIBZ
#include <zlib.h>
#endif
#include "job.h"
#include "capture.h"
#include "app_data.h"
#include "control.h"
#include "util.h"
//#include "sound.h"
#include "xtopnm.h"
#include "xtoxwd.h"
#include "xtopng.h"
#include "xtomng.h"
#include "xtojpg.h"
#include "xtoqtf.h"
#include "xtoffmpeg.h"
#include "frame.h"
#include "colors.h"
#include "codecs.h"

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <pango/pango-layout.h>
#include <gdk-pixbuf/gdk-pixdata.h>


static Job *job;
//extern GtkWidget window1;
extern xvCodec tCodecs[NUMCODECS];
extern xvFFormat tFFormats[NUMCAPS];

void job_dump ();

/*
 * HELPER FUNCTIONS
 */
Job *
job_new () {
    job = (Job *) malloc (sizeof (Job));
    if (!job) {
        perror ("job.c: job_new() bailed out ... malloc failed?!?");
        exit (1);
    }
    return (job);
}

void job_set_from_app_data (AppData *app, Display *disp, XWindowAttributes wa) {
    CapTypeOptions *cto;
    char file[PATH_MAX+1];
    
    // make sure we do have a job
    if ( job == NULL ) {
        perror ("job.c: job_set_from_app_data() has a NULL job ... this should never happen!");
        exit (1);
    }
    
    // switch sf or mf
    if (app->current_mode == 0) cto = &(app->single_frame);
    else cto = &(app->multi_frame);
    
    // get area to capture
    job->area = GetArea ();
    // various manual settings
    // need to have the flags set to smth. before job_set_file calls job_set_save_function
    // which will try to set the FLG_MULTI_IMAGE
    job->flags = app->flags;    // FIXME: will this work out? or do we need cto->flags?
    job_set_fps (cto->fps);

    job->state = VC_STOP;       // FIXME: better move this outta here?
    job->start_no = cto->start_no;
    job->pic_no = job->start_no;
    job->movie_no = 0;          // FIXME: make this configurable
    job->step = cto->step;
    //    int time_per_frame; /* time per frame in milli secs ... this is set in job_set_fps() */
    job->max_frames = cto->frames;
    job->max_time = cto->time;
    job->compress = cto->compress;
    job->quality = cto->quality;
    //	char open_flags[8];
    job->bpp = cto->bpp;
    //	int vid_dev;		/* file descriptor for video device */
    //	int snd_dev;		/* file descriptor for sound device */
    // FIXME: watch out for HAVE_FFMPEG_AUDIO
    if (cto->audioWanted == 0) job->flags &= ~FLG_REC_SOUND;
    else job->flags |= FLG_REC_SOUND;
    job->snd_rate = cto->sndrate;
    job->snd_smplsize = cto->sndsize;
    job->snd_channels = cto->sndchannels;
    job->mouseWanted = app->mouseWanted;
    job->video_dev = app->device;
    job->snd_device = app->snddev;
    
    job->get_colors = (void*(*)(XColor *, int))NULL;
    //	void *(*open)(char *, char *);
    //	int   (*write)(void *, unsigned int, unsigned int, void *);
    //	int   (*close)(void *);
    //	void  (*save)(FILE *, XImage *, struct _Job *);
    job->clean = (void(*)(Job *))NULL;
    
    job->target = cto->target;
    if (job->target <= 0) {
        // we can safely assume cto->filename is longer than 0 for the next bit because
        // with target == 0 it would have been set to a default value otherwise
        if (job->target == 0) job->target = xv_codec_get_target_from_filename (cto->file);
        // we assume job->target can never be == 0 now, because a sanity checking function
        // should have checked before if we have a valid specification for a target either
        // through target itself or the filename extension
        if (job->target <= 0) {
            fprintf (stderr, "Unrecoverable error while initializing job from app_data.\n");
            fprintf (stderr, "target is still '%i'. This should never happen.\n", job->target);
            fprintf (stderr, "Please contact the xvidcap project team.\n");
            exit (1);
        }
    } 
    
    job->targetCodec = cto->targetCodec;
    if (job->targetCodec <= 0) {
        if (job->targetCodec == 0) job->targetCodec = tFFormats[job->target].def_vid_codec;
        if (job->targetCodec < 0) {
            fprintf (stderr, "Unrecoverable error while initializing job from app_data.\n");
            fprintf (stderr, "targetCodec is still < 0. This should never happen.\n");
            fprintf (stderr, "Please contact the xvidcap project team.\n");
            job_dump ();
            exit (1);
        }
    } 
    job->au_targetCodec = cto->au_targetCodec;
    if (job->au_targetCodec <= 0) {
        if (job->au_targetCodec == 0) job->au_targetCodec = tFFormats[job->target].def_au_codec;
        // if 0 the format has no default audio codec. This should only be the case if the format
        // does not support audio
        if (job->au_targetCodec < 0) {
            fprintf (stderr, "Unrecoverable error while initializing job from app_data.\n");
            fprintf (stderr, "au_targetCodec is still < 0. This should never happen.\n");
            fprintf (stderr, "Please contact the xvidcap project team.\n");
            job_dump ();
            exit (1);
        }
    } 
    //	unsigned long crc;
    //	int ncolors;
    
    job->color_table = NULL;
    job->colors = NULL;
    job->win_attr = wa; 
    //	XRectangle *area;
    
    // the order of the following actions is key!
    // the default is to use the colormap of the root window
    job->ncolors = GetColors (disp, &job->win_attr, &job->colors);
    job_set_capture ();

    // set temporary filename if set to "ask-user"
    if (strlen (cto->file) < 1 ) {
        char *home = NULL;
        int pid;
        
        home = getenv ("HOME");
        pid = (int) getpid();
        
        snprintf (file, (PATH_MAX+1), "%s/.xvidcap-tmp.%i.%s", home, pid, xv_next_element (tFFormats[job->target].extensions));
    } else {
        snprintf (file, (PATH_MAX+1), "%s", cto->file);
    } 
    
    job_set_file (file);
    job_set_save_function (job->win_attr.visual, job->target);
    // get color table
    if (job->get_colors)
        job->color_table = (*job->get_colors)(job->colors, job->ncolors);
    
#ifdef DEBUG2
    printf ("job_set_from_add_data()\n");
#endif
//	job_dump();

}

/*
 * get a pointer to the static Job
 *
 */
Job *
job_ptr (void) {
    return (job);
}




/*
 * Select Function to use for processing the captured frame
 */
void
job_set_save_function (Visual *vis, int type) {
    switch (type) {
        case CAP_PNM:
            job->get_colors = PPMcolorTable;
            job->clean = PnmClean;
            job->flags &= ~FLG_MULTI_IMAGE;
            
            switch (vis->class) {
                case TrueColor:
                case DirectColor:
                    job->save = XImageToPPMC;
                    break;
                case PseudoColor:
                case StaticColor:
                    job->save = XImageToPPM8;
                    break;
                case StaticGray:
                case GrayScale:
                    printf ("Visual not supported\n");
                    job->save = XImageToPGM;
                    break;
            }
            break;
#ifdef HAVE_LIBPNG
        case CAP_PNG:
            job->get_colors = PNGcolorTable;
            job->clean = PngClean;
            job->flags &= ~FLG_MULTI_IMAGE;
            
            switch (vis->class) {
                case TrueColor:
                case DirectColor:
                    job->save = XImageToPNGC;
                    break;
                case PseudoColor:
                case StaticColor:
                case StaticGray:
                case GrayScale:
                    job->save = XImageToPNG8;
                    break;
            }
            break;
        case CAP_MNG:
            job->get_colors = PNGcolorTable;
            job->flags |= FLG_MULTI_IMAGE;
            job->clean = MngClean;
            
            switch (vis->class) {
                case TrueColor:
                case DirectColor:
                    job->save = XImageToMNGC;
                    break;
                case PseudoColor:
                case StaticColor:
                case StaticGray:
                case GrayScale:
                    job->save = XImageToMNG8;
                    break;
            }
            break;
#endif // HAVE_LIBPNG
#ifdef HAVE_LIBJPEG
        case CAP_JPG:
            job->get_colors = JPGcolorTable;
            job->clean = JpgClean;
            job->flags &= ~FLG_MULTI_IMAGE;
            
            switch (vis->class) {
                case TrueColor:
                case DirectColor:
                    job->save = XImageToJPGC;
                    break;
                case PseudoColor:
                case StaticColor:
                    job->save = XImageToJPG8;
                    break;
                case StaticGray:
                    job->save = XImageToJPGG;
                    break;
                case GrayScale:
                    job->save = XImageToJPGG;
                    printf ("Visual not supported!\n");
                    break;
            }
            break;
#endif // HAVE_LIBJPEG
#ifdef WITHGT
        case CAP_QTR:
        case CAP_QTJ:
        case CAP_QTP:
            job->clean = QtClean;
            job->flags |= FLG_MULTI_IMAGE;
            
            switch (vis->class) {
                case TrueColor:
                case DirectColor:
                    job->save = XImageToQTFC;
                    break;
                case StaticGray:
                case GrayScale:
                    job->save = XImageToQTFG;
                    break;
                default:
                    printf ("Visual not supported!\n");
                    exit (2);
                    break;
            }
            break;
#endif // WITHGT
#ifdef HAVE_LIBAVCODEC
#ifdef HAVE_LIBAVFORMAT
        case CAP_AVI:
        case CAP_MPG:
        case CAP_ASF:
        case CAP_FLV:
        case CAP_SWF:
            job->clean = FFMPEGClean;
            if (job->targetCodec == CODEC_NONE) {
                job->targetCodec = CODEC_MPEG1;
            }
            job->flags |= FLG_MULTI_IMAGE;
            job->get_colors = FFMPEGcolorTable;
            
            /*
             * make sure we have even width and height for ffmpeg
             */
/*
 * this should be caught by job-validate
            {
                Boolean changed = FALSE;
                
                if (job->flags & FLG_RUN_VERBOSE) {
                    fprintf (stderr, "job.job_set_save_function(): Original dimensions: %i * %i\n", job->area->width, job->area->height);
                }
                if ((job->area->width % 2) > 0 ) {
                    job->area->width--;
                    changed = TRUE;
                }
                if ((job->area->height % 2) > 0) {
                    job->area->height--;
                    changed = TRUE;
                }
                if ( job->area->width < 20 ) {
                    job->area->width = 20;
                    changed = TRUE;
                }
                if ( job->area->height < 20 ) {
                    job->area->height = 20;
                    changed = TRUE;
                }
                
                if (changed) {
                    XVC_ChangeFrame (job->area->x, job->area->y, job->area->width, job->area->height, FALSE);
                    if (job->flags & FLG_RUN_VERBOSE) {
                        fprintf (stderr, "job.job_set_save_function(): Modified dimensions: %i * %i\n", job->area->width, job->area->height);
                    }
                }
            } */
            
            job->save = XImageToFFMPEG;
            break;
#endif // HAVE_LIBAVFORMAT
#endif // HAVE_LIBAVCODEC
        default:	/* XWD */
            job->save = XImageToXWD;
            job->get_colors = XWDcolorTable;
            job->flags &= ~FLG_MULTI_IMAGE;
            job->clean = NULL;
            break;
    }
}


/*
 */
void
job_set_fps (int fps) {
// this is checked in app->validate
//    if (fps <= 0)
//        fps = 0.00001; /* min value */
    job->time_per_frame = (int) ( 1000 / (fps / (float) 100));
    job->fps = fps;
}


#ifdef HAVE_FFMPEG_AUDIO
/*
 * set and check some parameters for the sound device
 */
 
int
job_set_sound_dev (char *snd, int rate, int size, int channels) {
    extern int errno;
    struct stat statbuf;
    int stat_ret;
    
    job->snd_rate = rate;
    job->snd_smplsize = size;
    job->snd_channels = channels;
    job->snd_device = snd;
    
    if (job->flags & FLG_REC_SOUND) {
        if (strcmp (snd, "-")!=0) {
            stat_ret = stat (snd, &statbuf);
            
            if ( stat_ret != 0 ) {
                switch (errno) {
                    case EACCES:
                        fprintf (stderr, "Insufficient permission to access sound input from %s\n", snd);
                        fprintf (stderr, "Sound disabled!\n");
                        job->flags &= ~FLG_REC_SOUND;
                        break;
                    default:
                        fprintf (stderr, "Error accessing sound input from %s\n", snd);
                        fprintf (stderr, "Sound disabled!\n");
                        job->flags &= ~FLG_REC_SOUND;
                        break;
                }
            }
        }
    } // end if soundWanted
    return 0;
}
#endif // HAVE_FFMPEG_AUDIO



/*
 */
void
job_set_file (char *file) {
    
    if (!file)
        return;
    /*
    if ( job->target < 0 ) job->target = xv_codec_get_target_from_filename(file);
    if (job->target < 0) {
        fprintf(stderr, "file format for file %s is not supported with this binary. Aborting!\n", file);
        exit(1);
    }
    else if (job->target == 0) {
        printf("unknown file format requested through file %s. Aborting!\n", file);
        exit(1);
    }
     */
    
#ifdef HAVE_LIBZ
    if (job->compress && !(
#ifdef HAVE_LIBPNG
    (job->target == CAP_PNG) || (job->target == CAP_MNG) ||
#ifdef WITHGT
    (job->target == CAP_QTP) ||
#endif //WITHGT
#endif //HAVE_LIBPNG
    (job->target >= CAP_FFM))) {
        int len = strlen (file);
        
        job->open = (void*(*)(char*, char*))gzopen;
        job->close= (int(*)(void*))gzclose;
        job->write = zwrite;
        if (job->file && (job->file != file))
            free (job->file);
        if (!strstr (file, ".gz")) {
            job->file = (char *) malloc (len + 4 + 1);
            sprintf (job->file, "%s.gz", file);
        } else {
            job->file = strdup (file);
        }
        sprintf (job->open_flags, "wb%d", job->compress);
    }
#endif
    job->open = (void*(*)(char*, char*))fopen;
    job->close= (int(*)(void*))fclose;
#ifdef WITHGT
#ifdef HAVE_LIBJPEG
    if (job->target == CAP_QTJ) {
        job->write = (int(*)(void*, unsigned int, unsigned int, void *))jwrite;
    } else
#endif //HAVE_LIBJPEG
#ifdef HAVE_LIBPNG
        if (job->target == CAP_QTP) {
            job->write = (int(*)(void*, unsigned int, unsigned int, void *))xv_pwrite;
        } else
#endif //HAVE_LIBPNG
#endif //WITHGT
{
    job->write = (int(*)(void*, unsigned int, unsigned int, void *))fwrite;
}
job->file = strdup (file);

if (strstr (file, ".gz")) {
    int len = strlen (file);
    job->file[len-3] = '\0';
}
sprintf (job->open_flags, "wb");

#ifdef HasBTTV
    if (job->flags & FLG_USE_V4L) {
        job->capture = TCbCaptureV4L;
    }
#endif
}



/*
 */
void
job_set_compression (int comp) {
    if (comp < 0)
        comp = 0;
    else if (comp > 9)
        comp = 9;
    job->compress = comp;
}


/*
 * find the correct capture function
 */
void
job_set_capture (void) {
    int input = job->flags & FLG_SOURCE;
    
    switch (input) {
        case FLG_USE_SHM:
            job->capture = TCbCaptureSHM;
            break;
        case FLG_USE_DGA:
            job->capture = TCbCaptureDGA;
            break;
        case FLG_USE_V4L:
            job->capture = TCbCaptureV4L;
            break;
        default:
            job->capture = TCbCaptureX11;
            break;
    }
}

/*
 */
void
job_set_quality (int quality) {
    if (quality < 1)
        quality = 1;
    else if (quality > 100)
        quality = 100;
    job->quality = quality;
}



/*
 * dump job settings
 */
void job_dump () {
    printf ("fps		= %.2f\n", (float) (job->fps / 100));
    printf ("file		= %s\n", job->file);
    printf ("flags		= %i\n", job->flags);
    printf ("state		= %i\n", job->state);
    printf ("start_no	= %i\n", job->start_no);
    printf ("pic_no		= %i\n", job->pic_no);
    printf ("movie_no	= %i\n", job->movie_no);
    printf ("step		= %i\n", job->step);
    printf ("time_per_frame	= %i\n", job->time_per_frame);
    printf ("max_frames	= %i\n", job->max_frames);
    printf ("max_time	= %i\n", job->max_time);
    printf ("compress	= %i\n", job->compress);
    printf ("quality		= %i\n", job->quality);
    printf ("open_flags	= %s\n", strdup (job->open_flags));
    printf ("bpp		= %i\n", job->bpp);
    printf ("vid_dev		= %i\n", job->vid_dev);
    printf ("snd_dev		= %i\n", job->snd_dev);
    printf ("snd_rate	= %i\n", job->snd_rate);
    printf ("snd_smplsize	= %i\n", job->snd_smplsize);
    printf ("snd_channels	= %i\n", job->snd_channels);
    printf ("mouseWanted	= %i\n", job->mouseWanted);
    printf ("video_dev	= %s\n", job->video_dev);
    printf ("snd_device	= %s\n", job->snd_device);
    
    printf ("get_colors	= %p\n", job->get_colors);
    printf ("open		= %p\n", job->open);
    printf ("write		= %p\n", job->write);
    printf ("close		= %p\n", job->close);
    printf ("save		= %p\n", job->save);
    printf ("clean		= %p\n", job->clean);
    printf ("capture		= %p\n", job->capture);
    
    printf ("target		= %i\n", job->target);
    printf ("targetCodec	= %i\n", job->targetCodec);
    //	unsigned long crc;
    printf ("ncolors		= %i\n", job->ncolors);
    
    printf ("color_table	= %p\n", job->color_table);
    printf ("colors		= %p\n", job->colors);
    printf ("win_attr (w/h/x/y)	= %i/%i/%i/%i\n", job->win_attr.width, job->win_attr.height,
    job->win_attr.x, job->win_attr.y);
    printf ("area (w/h/x/y)		= %i/%i/%i/%i\n", job->area->width, job->area->height,
    job->area->x, job->area->y);
    
}


/*
 * validate job
 * some things, can only be set after the app_data things have become a job
 * these things include e.g. flags and capture size
 */
void job_validate () {
    // unset autocontinue unless we capture to movie and file is mutable
    if ( job->flags & FLG_AUTO_CONTINUE && ( (! is_filename_mutable (job->file)) ||
    (job->flags & FLG_MULTI_IMAGE) == 0 ) ) {
        job->flags &= ~FLG_AUTO_CONTINUE;
        printf ("Output not a video file or no counter in filename\nDisabling autocontinue!\n");
    }
    
#ifdef HAVE_LIBAVCODEC
#ifdef HAVE_LIBAVFORMAT
    /*
     * make sure we have even width and height for ffmpeg
     */
    if (job->target >= CAP_FFM) {
        Boolean changed = FALSE;
        
        if (job->flags & FLG_RUN_VERBOSE) {
            fprintf (stderr, "control.InitControl(): Original dimensions: %i * %i\n", job->area->width, job->area->height);
        }
        if ((job->area->width % 2) > 0 ) {
            job->area->width--;
            changed = TRUE;
        }
        if ((job->area->height % 2) > 0) {
            job->area->height--;
            changed = TRUE;
        }
        if ( job->win_attr.width < 26 ) {
            job->win_attr.width = 26;
            changed = TRUE;
        }
        if ( job->win_attr.height < 26 ) {
            job->win_attr.height = 26;
            changed = TRUE;
        }
        
        if (changed) {
            XVC_ChangeFrame (job->area->x, job->area->y, job->area->width, job->area->height, FALSE);
            if (job->flags & FLG_RUN_VERBOSE) {
                fprintf (stderr, "control.InitControl(): Modified dimensions: %i * %i\n", job->area->width, job->area->height);
            }
        }
        
        /*
         * reset targetCodec to default for file format if unallowed codec is selected
         */
/*
 * this should be caught earlier ...
        if ( ! xv_is_element ( tFFormats[job->target].allowed_vid_codecs, tCodecs[job->targetCodec].name ) ) {
            fprintf (stderr, "%s is not an allowed video codec for file format %s\n",
            tCodecs[job->targetCodec].name, tFFormats[job->target].name);
            fprintf (stderr, "Resetting to default codec for file format: %s\n",
            tCodecs[tFFormats[job->target].def_vid_codec].name );
            job->targetCodec = tFFormats[job->target].def_vid_codec;
        }
*/


    }
    
#endif // HAVE_LIBAVFORMAT
#endif // HAVE_LIBAVCODEC




}



